/* Portions Copyright (c) 2005 Financial Management Service and specially
 *  designated fiscal or financial agents
 * Portions Copyright (c) 2005 RAF Technology, Inc.
 *
 * This software shall not be used, reproduced, sold or disclosed in whole or
 * in part without the express written permission of RAF Technology, Inc.,
 * except to the extent Financial Management Service and specially designated
 * fiscal or financial agents have been granted specific rights under license. */
package com.iwombat.foundation;

import java.io.Serializable;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Random;

import com.iwombat.foundation.log.Log;
import com.iwombat.util.HexStringUtil;

/**
 * Universal object identifier. <br>
 * 16-byte Object identifier for use by all persistant objects based on DCE standard <br>
 * <ul>
 * <li>byte 00-03: low time comonent
 * <li>byte 04-05: mid time component
 * <li>byte 06-07: high time component and version
 * <li>byte 08: sequence number high byte and variant
 * <li>byte 09: sequence number low byte
 * <li>byte 10-15: MAC address - passed in from System property
 * </ul>
 * <br>
 * 
 * Implementation notes: The DCE spec calls for MacAddress as part of the UUID algorythm. Since java supplies no
 * mechanism for this ObjectId looks for a MACADDR system property consisting of a hex string of six bytes ( a 12
 * character hex string - no delimiters and no 0x token ). Failing to find said system property ObjectId generates a
 * random MACADDR and uses the last octet of the ip address as the last octet in the MACADDR to guarantee unequeness on
 * the same subnet. (If localhost is NOT 127.0.0.1)
 * 
 * In order to be more friendly to bean serializers for both XML serialization in jdk1.4 and WSDL bean serializers in
 * apache axis, the field id, and its bean accessor and mutator are now strings. setId() keeps the byte[] binOid (useful
 * for RDBMS persistence) and String id (useful for XML serialization) in sync.
 * 
 * @author bob
 * @deprecated please use UUID and UUIDFactory implementations
 */
public final class ObjectId implements Identifier, Serializable {

    private static final long serialVersionUID = 1;
    
    // length of the ObjectId byte array
    private static final short OID_LENGTH = 16;

    // byte array containing the generated id
    private byte[] binOid = new byte[OID_LENGTH];

    private String id = null;

    // length to shift by one byte
    private static final short BYTE_SHIFT = 8;

//    private static final byte ONE_TWENTY_SEVEN = (byte) 0x7F;

    // byte and nibble masks
    private static final short LOW_BYTE_MASK = (short) ((short) (1 << BYTE_SHIFT) - 1);

//    private static final short HIGH_NIBBLE_MASK = (short) 0xf0;

    private static final short LOW_NIBBLE_MASK = (short) 0x0f;

    // maximum sequence number (four bytes)
    // leave room on high byte to stuff a version number
    private static final int MAX_SEQUENCE_NUMBER = 0x4000;

    // static sequence number
    private static int sequenceNumber = 0;

    // static MAC address
    private static byte[] macAddr;

    private static final short MACADDR_LENGTH = 6;

    // random number generator
    private static Random rand = new Random(System.currentTimeMillis());

    // last time encountered
    private static long lastTimestamp = 0;

    // Time masks and offsets
    private static final long LOW_TIME_MASK = 0x00000000ffffffffL;

    private static final long MID_TIME_MASK = 0x0000ffff00000000L;

    private static final long HIGH_TIME_MASK = 0x0fff000000000000L;

    private static final short MID_TIME_OFFSET = BYTE_SHIFT * 4;

    private static final short HIGH_TIME_OFFSET = BYTE_SHIFT * 6;

    // Version and variant
    private static final byte VERSION = (byte) 0x10;

    private static final byte VARIANT = (byte) 0x80;

    // generate the number of 100 nanosecond intervals between 10/15/1582
    // and 1/1/1970 - the beginning of the Java clock
    private static final long NANOSECONDS_PER_DAY = 86400 * 10000000;

    private static final long TIME_OFFSET = NANOSECONDS_PER_DAY * ((365 * (1970 - 1583)) // days per year
            + (24 * 3) // leap days per century
            + (5) // leap days 1563-1600 inclusive
            + (17) // leap days 20th century
            + (17 + 30 + 31) // remaining days in 1582
    - (12) // calendar reset in 9/1752
            );

    // set the value of the MAC address at class initialization time
    static {
        // get the machine's ip address once per JVM invocation of the class
        String addrString = System.getProperty("MACADDR");
        try {
            macAddr = HexStringUtil.bytesFromHexString(addrString);
        } catch (RuntimeException e) {
            macAddr = getDummyMACADDR();
        }

        // make sure we got a valid MACAddr length
        // we may just want to just die here instead
        if (macAddr.length < MACADDR_LENGTH) {
            System.err.println("StartupError: MACADDR not defined properly in System.properties");
            macAddr = getDummyMACADDR();
        }

        // seed the sequence number with a random
        sequenceNumber = rand.nextInt();
        sequenceNumber %= MAX_SEQUENCE_NUMBER;
    }

    /**
     * increment the sequence number -- synchronized for your protection
     */
    private static synchronized void incrementSequence() {
        ++sequenceNumber;
        sequenceNumber %= MAX_SEQUENCE_NUMBER;
    }

    private static synchronized void setTimeStamp(long time) {
        lastTimestamp = time;
    }

    /**
     * get a new time if not incremented since the last OID was generated we increment the sequence number
     */
    private long getTime() {
        long time = TIME_OFFSET + (1000 * System.currentTimeMillis());
        if (!(time > lastTimestamp)) {
            incrementSequence();
        } else {
            setTimeStamp(time);
        }

        return time;
    }

    /**
     * convienience method puts a long into the id buffer little endian style - decending
     * 
     * @param arrayPosition
     *            beginning position
     * @param numBytes
     */
    private void stuffOidWithLong(byte[] tempOid, long l, int arrayPosition, int numBytes) {
        int index = (arrayPosition + numBytes) - 1;
        do {
            tempOid[index] = (byte) (l & LOW_BYTE_MASK);
            l >>>= BYTE_SHIFT;
        } while (index-- > arrayPosition);
    }

    private static byte[] getDummyMACADDR() {
        byte[] dummy = new byte[6];
        rand.nextBytes(dummy);

        // set the broadcast bit so we know its a dummy
        dummy[0] |= (byte) 0x80;

        // put the last octet of the ip address in index 5 for some
        // guarantee of uniqueness on the same subnet
        // unless the localhost is set to 127.0.0.1
        try {
            InetAddress localhost = InetAddress.getLocalHost();
            byte[] address = localhost.getAddress();
            if ((byte) 0x7F != address[0]) {
                dummy[dummy.length - 1] = address[address.length - 1];
            }
        } catch (UnknownHostException e) {
            // do nothing here and settle with the random numbers
            Log.info(ObjectId.class, "Using random numbers to generate ids.");
        }
        return dummy;
    }

    /**
     * Generate an oid from scratch.
     */
    private byte[] createOid() {
        long naughtyBits = 0;
        byte[] localOid = new byte[OID_LENGTH];

        // get since epoch and convert it to a byte array
        long time = getTime();

        // stuff the low time component into id
        naughtyBits = time & LOW_TIME_MASK;
        stuffOidWithLong(localOid, naughtyBits, 0, 4);

        // stuff the mid time component into id
        naughtyBits = (time & MID_TIME_MASK) >>> MID_TIME_OFFSET;
        stuffOidWithLong(localOid, naughtyBits, 4, 2);

        // stuff the high time component into id
        naughtyBits = (time & HIGH_TIME_MASK) >>> HIGH_TIME_OFFSET;
        stuffOidWithLong(localOid, naughtyBits, 6, 2);

        naughtyBits = sequenceNumber;
        stuffOidWithLong(localOid, naughtyBits, 8, 2);

        // the machine's macaddr - do it by hand
        System.arraycopy(macAddr, 0, localOid, 10, MACADDR_LENGTH);

        // put the version into position 6
        localOid[6] &= LOW_NIBBLE_MASK;
        localOid[6] |= VERSION;

        // put the variant into position 8
        localOid[8] &= LOW_NIBBLE_MASK;
        localOid[8] |= VARIANT;

        return localOid;
    }

    /**
     * Generic Constructor
     */
    public ObjectId() {
        this.setId(createOid());
    }

    /**
     * Constructor with known ObjectId value. Throws an Exception if the argument is not the correct length.
     * 
     * @param binoid the given value
     * @throws IllegalArgumentException if null or bad length
     */
    public ObjectId(byte[] binoid) {
        this.setId(binoid);
    }

    /**
     * Constructor with known ObjectId value. Throws an Exception if the argument is not the correct length.
     * 
     * @param hexString the given value
     * @throws IllegalArgumentException if null or bad length
     */
    public ObjectId(String hexString) {
        this.setId(hexString);
    }

    /**
     * Accessor for the binary value of this ObjectId.
     * 
     * @return byte-aray containing the value of the ObjectId
     */
    public byte[] getValue() {
        return binOid;
    }

    /**
     * This is a bean pattern setter for use with BeanSearializers in Axis framework ONLY. You should never use this
     * setter externally except for deserializers.
     * 
     * @param newBinOid a new id
     */
    public void setId(byte[] newBinOid) {
        if (newBinOid == null) {
            throw new IllegalArgumentException("argument null value");
        }
        this.setStringAndByteId(newBinOid, HexStringUtil.stringFromBytes(newBinOid));
    }

    /**
     * This is a bean pattern setter for use with BeanSearializers in Axis framework ONLY. You should never use this
     * setter externally except for deserializers.
     * @param hexString a new hex String id
     */
    public void setId(String hexString) {
        if (hexString == null) {
            throw new IllegalArgumentException("argument null value");
        }

        this.setStringAndByteId(HexStringUtil.bytesFromHexString(hexString), hexString);
    }

    private void setStringAndByteId(byte[] newBinOid, String newHexString) {
        binOid = newBinOid;
        id = newHexString;
    }

    /**
     * Bean pattern getter for use with BeanSearializers in Axis framework 
     * @return byte value
     */
    public String getId() {
        return id;
    }

    /**
     * Gets a hex string representation of this ObjectId. Each byte in the binary value is represented by two hex
     * digits.
     * 
     * @return String representation of the ObjectId
     */
    public String toHexString() {
        return getId();
    }

    /**
     * Compares for equality.
     * 
     * @return true if o is an ObjectId, and it's value equal.
     * @param obj The object to which this id should be compared.
     */
    public boolean equals(Object obj) {
        if (!(obj instanceof ObjectId)) {
            return false;
        }

        byte[] otherBinaryValue = ((ObjectId) obj).getValue();
        byte[] myBinaryValue = this.getValue();

        for (int counter = 0; counter < OID_LENGTH; counter++) {
            if (otherBinaryValue[counter] != myBinaryValue[counter]) {
                return false;
            }
        }
        return true;
    }

    /**
     * Supports the java obect hashCode - lamely.
     * 
     * @return hashcode.
     */
    public int hashCode() {
        return this.toHexString().hashCode();
    }

    /**
     * Print the object id out as a human-readable hex string.
     * @return The value of the object ID as a hex string.
     */
    public String toString() {
        return this.toHexString();
    }
}

