// (c) Copyright 2001-2002 iwombat.com all rights reserved

package com.iwombat.foundation.persist;

/**
 * Class AbstractResultSetToBeanMapper
 * Description: Abstract supporting class to map source result set items to target beans
 *
 * <br><br>
 * Data types supported:<br>
 * com.iwombat.Identifier<br>
 * java.lang.String<br>
 * java.lang.Long<br>
 * java.langFloat<br>
 * java.util.Date<br>
 * java.lang.Boolean<br>
 * com.iwombat.ByteArray<br>
 * <br><br>
 * Note that language primitives are not supported because they cannot reflect the concept of 'Null'.
 * <br>
 * @author  $Author: brentp $
 * @version $Revision: 1.1 $
 * 
 * <br>
 * <dl><dt><b>Date:</b><dd>$Date: 2002/11/18 16:04:39 $</dd></dl>
 */
import java.lang.reflect.Method;

import java.sql.*;

import java.util.*;

import com.iwombat.util.*;

import com.iwombat.foundation.*;

public abstract class AbstractResultSetToBeanMapper implements ResultSetToBeanMapper{
    
    private Object targetBean;
    private Map mutatorMethods;
    private ResultSet sourceResultSet;
    
    /**
     * retrieves the source result set to operate against
     *
     * @return the currently set ResultSet
     */
    public final ResultSet getSourceResultSet() {
        return this.sourceResultSet;
    }

    /**
     * sets the source result set to operate against
     *
     * @param the ResultSet to use
     */
    public final void setSourceResultSet(ResultSet resultSet) {
        this.sourceResultSet=resultSet;
    }

    /**
     * sets the target bean to map result set values into
     *
     * @param the target bean to use
     */
    public final void setTargetBean(Object targetBean){
        this.targetBean=targetBean;
        //invalidate the current map of mutator methods for this target bean
        this.mutatorMethods=null;
    }
    
    /**
     * retrives the currently set target bean to map result set values into
     *
     * @return the currently set target bean
     */
    public final Object getTargetBean(){
        return this.targetBean;
    }
    
    /**
     * retrieves a unique Map of bean property mutator Method objects (setXXX, setYYY, setZZZ, etc)
     * <br><br>
     * The map will contain only bean property mutator methods that contain a single parameter, and only returns
     * the first mutator method of a kind; it will not recognize overloaded mutator methods.
     * <br><br>
     * The key of the map will contain the name of the method, and the value referenced by the key will contain the corresponding
     * Method Object.
     *
     * @return the Map of mutator methods for the target bean
     */
    protected final Map getMutatorMethods(){
        
        if (null!=mutatorMethods){
            return mutatorMethods;
        }
        else{
            //load it up
            Object bean = getTargetBean();
            if (null==bean) throw new IllegalStateException("Target bean is not defined");


            this.mutatorMethods=new HashMap();

            Method methods[]=bean.getClass().getMethods();

            for (int i=0;i<methods.length;i++){
                Method m=methods[i];
                String name=m.getName();
                //only methods starting with "set" with single parameters are valid
                if (name.startsWith("set") && m.getParameterTypes().length==1){
                    //we don't support overloaded setters. sorry
                    if (!this.mutatorMethods.containsKey(name)){
                        this.mutatorMethods.put(name, m);
                    }
                }
            }
            return this.mutatorMethods;
        }
    }
    
    /**
     * Retrieves the Bean property mutator 
     * Method object corresponding to the specified bean 
     * property name for the currently set target Bean.
     *
     * @return the Method object corresponding to the specified bean property name
     */
    protected final Method locateMutatorMethod(String beanPropertyName){
        //find the bean property setter for the specified target bean / bean property name
        String methodName = BeanUtil.createMutatorMethodName(beanPropertyName);
        Method m = (Method)this.getMutatorMethods().get(methodName);
        
        if (m==null) throw new PersistenceException("Unable to find mutator Method for bean property '"+beanPropertyName+"'.");
        
        return m;
    }
    
    /**
     * invokes the specified Bean property mutator method with the specified value on the currently
     * set target Bean
     *
     * @param the mutator method to invoke
     * @param the value to be specified in the mutator method
     */
    protected final void invokeBeanPropertySetter(Method mutatorMethod, Object value){

        Object bean = this.getTargetBean();
        if (null==bean) throw new IllegalStateException("Target Bean is not defined");
        
        try{
            mutatorMethod.invoke(bean,new Object[]{value});
        }
        catch (Throwable t){
            throw new PersistenceException("Unable to call set(BeanPropertyName) method: '" + mutatorMethod.getName() + "', for target bean '" + targetBean.getClass().getName()+"':\nt.getClass()=[" + t.getClass() + "] t.getMessage()=[" + t.getMessage() + "]", t);
        }
    }
    
    /**
     * Sets the currently set bean property from the value in the currently set result set
     * based on the specified query result to bean configuration.
     *
     * @param the query result to bean configuration to use for this operation
     */
    public final void setBeanProperty(QueryResultToBeanMapItem queryResultToBeanConfig) {
        
        Object targetBean=this.getTargetBean();
        if (null==targetBean) throw new IllegalStateException("target bean is not defined");
        
        ResultSet sourceResultSet=this.getSourceResultSet();
        if (null==sourceResultSet) throw new IllegalStateException("result set is not defined");

        String columnName = queryResultToBeanConfig.getColumnName();
        String beanPropertyName=queryResultToBeanConfig.getBeanPropertyName();
        Method mutatorMethod=queryResultToBeanConfig.getMutatorMethod();

        //cache the mutator method in the specified QueryResultToBeanMapItem - it won't need to
        //change for the lifetime of the running Java VM
        if (mutatorMethod==null){
            mutatorMethod=locateMutatorMethod(queryResultToBeanConfig.getBeanPropertyName());
            queryResultToBeanConfig.setMutatorMethod(mutatorMethod);
        }
        
        if (mutatorMethod==null){
            throw new PersistenceException("Did not find expected set method for bean property name '" + beanPropertyName + "'; target bean '" + targetBean.getClass().getName() +"'");
        }

        //we expect one parameter type to be defined
        Class beanPropertySetClass=mutatorMethod.getParameterTypes()[0];

        //test for one of the supported data types
        try{
            if (beanPropertySetClass==Identifier.class){
                setIdentifierBeanProperty(mutatorMethod, columnName);
            }
            else if (beanPropertySetClass==String.class){
                setStringBeanProperty(mutatorMethod, columnName);
            }
            else if (beanPropertySetClass==Long.class){
                setLongBeanProperty(mutatorMethod, columnName);
            }
            else if (beanPropertySetClass==Float.class){
                setFloatBeanProperty(mutatorMethod, columnName);
            }
            else if (beanPropertySetClass==Boolean.class){
                setBooleanBeanProperty(mutatorMethod, columnName);
            }
            else if (beanPropertySetClass==java.util.Date.class){
                setDateBeanProperty(mutatorMethod,columnName);
            }
            else if (beanPropertySetClass==ByteArray.class){
                setByteArrayBeanProperty(mutatorMethod,columnName);
            }
            else throw new PersistenceException("Unsupported bean property class: " + beanPropertySetClass.getName());
        }
        catch (java.sql.SQLException e){
            throw new PersistenceException("Unable to set bean property for target bean '" + targetBean.getClass().getName()+"'; Method '"+mutatorMethod.getName()+"'; result set column name '" + columnName +"'");
        }

    }
    
}
