// (c) Copyright 2001-2002 iwombat.com all rights reserved
package com.iwombat.foundation.persist;

import com.iwombat.foundation.environment.*;

/**
 * Retrieves a RdbmsSDAOOperationConfig based on the specified data access object, the persistable operation key, and the persistence 
 * technology key.
 *
 * The RdbmsDAOOperationConfig is stored as a serialized object (using 1.4 XML serialization), and is stored in the configuration base
 * directory under a sub directory named by the persistence technology key.
 * <br><br>
 * the RdbmsDAOOperationConfig contains all of the possible DAO operations for the specified data access object; the appropriate operation 
 * config is identified by the specified persistable operation key.
 * <br><br>
 * Example:
 * <br><br>
 * sample configuration:
 * <br><br>
 * /configBase/mssql/com.foobar.persistence.FooRdbmsDAOImpl.xml
 * /configBase/mssql/com.foobar.persistence.BarRdbmsDAOImpl.xml
 * /configBase/mysql/com.foobar.persistence.FooRdbmsDAOImpl.xml
 * /configBase/mysql/com.foobar.persistence.BarRdbmsDAOImpl.xml
 * <br><br>
 * Specifying the input of:
 * com.foobar.persistence.FooRdmbsDAOImpl, "getFooById", "mysql"<br>
 * would return the operation configuration for the operation "getFooById" in the /configBase/mysql/com.foobar.persistence.FooRdbmsDAOImpl.xml file.
 *
 * @author  $Author: bob $
 * @version $Revision: 1.4 $
 * 
 * <br>
 * <dl><dt><b>Date:</b><dd>$Date: 2002/11/26 00:21:58 $</dd></dl>
 */
public class RdbmsDAOOperationConfigLocator {
    
    /** force singleton */
    private RdbmsDAOOperationConfigLocator() {
    }

    public static RdbmsDAOOperationConfig getOperationConfig(
    DAO dataAccessObject,
    String persistableOperationKey,
    String persistenceTechnologyKey){
        return getOperationConfig(dataAccessObject.getClass(), persistableOperationKey, persistenceTechnologyKey);
    }
    /**
     * Retrieves the operation config specified by the data access object, the persistable operation key, and the persistence technology key
     * 
     * @param the data access object for the DAO Operation
     * @param the persistable operation key mapped within the operation config
     * @param the persistence technology key 
     */ 
    
    public static RdbmsDAOOperationConfig getOperationConfig(
    Class dataAccessObject,
    String persistableOperationKey,
    String persistenceTechnologyKey){
        
        
        if (dataAccessObject==null) throw new IllegalArgumentException("the DAO specified was null!");
        if (persistableOperationKey==null) throw new IllegalArgumentException("the persistable operation key specified was null!");
        if (persistenceTechnologyKey==null) throw new IllegalArgumentException("the persistence technology key specified was null!");
        
        String daoClassname = dataAccessObject.getName();
        
        RdbmsDAOOperationConfig operationConfig=null;
        
        while (null==operationConfig){
            try{
            operationConfig=(RdbmsDAOOperationConfig)ConfigurationLocator.getConfiguration(daoClassname, persistableOperationKey);
            if (null==operationConfig) throw new IllegalArgumentException("operationConfig not found for DAO '" + daoClassname + "'; operation key '" + persistableOperationKey + "'");
            }
            catch (EnvironmentException e){
                
                String operationConfigResourcePath=EnvironmentProperties.getProperty(daoClassname);
                
                //the operation config resource path is not yet in the environment properties
                if (null==operationConfigResourcePath){
                    
                    operationConfigResourcePath=getOperationConfigResourcePath(daoClassname,persistenceTechnologyKey);
                    
                    //set it in the environment properties
                    EnvironmentProperties.getProperties().setProperty(daoClassname, operationConfigResourcePath);
                    
                }
                //so the resource path exists yet the operation config wasn't loaded. something's fishy here.
                else{
                    throw new PersistenceConfigException("Unable to load Rdbms Operation config for DAO '" + daoClassname + "'; operation key '" + persistableOperationKey + "'; operation resource path '" + operationConfigResourcePath + "'; ("+e.getClass().getName()+") " + e.getMessage());
                }
            }
        }
            
    return operationConfig;
    }
    
    private static String getOperationConfigResourcePath(
                            String daoClassname,
                            String persistenceTechnologyKey){
     
        StringBuffer path=new StringBuffer();
        
        path.append(persistenceTechnologyKey);
        path.append("/");
        path.append(daoClassname);
        path.append(".xml");

        return path.toString();
    }
    
}