// (c) Copyright 2001-2002 iwombat.com all rights reserved
package com.iwombat.foundation.persist;

import java.sql.*;

import java.lang.reflect.*;

import java.util.*;

import com.iwombat.foundation.*;
import com.iwombat.foundation.environment.ConfigurationLocator;

import com.iwombat.util.*;

/**
 * Coordinates the execution of persistence operations
 *
 * @author  $Author: bob $
 * @version $Revision: 1.9 $
 * 
 * <br>
 * <dl><dt><b>Date:</b><dd>$Date: 2002/11/27 18:42:57 $</dd></dl>
 */

public class RdbmsDAOOperationExecutor {
    
    
    /** force singleton */
    private RdbmsDAOOperationExecutor() {
    }
    
    /**
     * executes an update-type operation, without transactional support
     *
     * @param the domain object to operate against (supplies input values)
     * @param the operation configuration to use
     * @param the SQL persistence to use
     * @return the result of the operation, if any
     */
    public static Object executeUpdateOperation(
    DomainObject domainObject,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence){
        
        return executeUpdateOperation(domainObject,opConfig,sqlPersistence,null);
    }
    
    /**
     * executes an update-type operation, supporting transactions
     *
     * @param the domain object to operate against (supplies input values)
     * @param the operation configuration to use
     * @param the SQL persistence to use
     * @param the transaction token to be used in this operation
     * @return the result of the operation, if any
     */
    public static Object executeUpdateOperation(
    DomainObject domainObject,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence,
    TransactionToken transactionToken){
        
        return executeOperation(domainObject,domainObject.getClass(),opConfig,sqlPersistence,transactionToken);
    }
    
    /**
     * executes an get-item type operation, without transactional support
     *
     * @param the input bean to use for this operation (typically a ParameterBean)
     * @param the domain object class to write results into
     * @param the operation configuration to use
     * @param the SQL persistence to use
     * @return the domain object created by this operation
     */

    public static DomainObject executeFetchItemOperation(
    Object inputBean,
    Class targetDomainObjectClass,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence){
        
        return executeFetchItemOperation(inputBean, targetDomainObjectClass, opConfig, sqlPersistence,null);
    }
    
    /**
     * executes an get-item type operation, supporting transactions
     *
     * @param the input bean to use for this operation (typically a ParameterBean)
     * @param the domain object class to write results into
     * @param the operation configuration to use
     * @param the SQL persistence to use
     * @param the transaction token to be used in this operation
     * @return the result of this operation
     */
    public static DomainObject executeFetchItemOperation(
    Object inputBean,
    Class targetDomainObjectClass,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence,
    TransactionToken transactionToken){
        
        
        Object result=executeOperation(inputBean,targetDomainObjectClass,opConfig,sqlPersistence,transactionToken);
        
        if (result==null) return null;
        
        if (result instanceof DomainObject){
            return (DomainObject)result;
        }
        else {
            throw new PersistenceException("Expected DomainObject to be returned as operation result: received '"
                                           + result.getClass().getName() + "' instead.");
        }
    }
    
    
    
    /**
     * executes an get list type operation, without transactional support
     *
     * @param the input bean to use for this operation (typically a ParameterBean)
     * @param the domain object class to create instances of and write item results into
     * @param the operation configuration to use
     * @param the SQL persistence to use
     * @return the list of domain objects created by this operation
     */
    public static Collection executeFetchListOperation(
    Object inputBean,
    Class targetDomainObjectClass,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence){
        
        return executeFetchListOperation(inputBean, targetDomainObjectClass, opConfig, sqlPersistence, null);
    }
    
    /**
     * executes an get list type operation, supporting transactions
     *
     * @param the input bean to use for this operation (typically a ParameterBean)
     * @param the domain object class to create instances of and write item results into
     * @param the operation configuration to use
     * @param the SQL persistence to use
     * @return the list of domain objects created by this operation
     */
    public static Collection executeFetchListOperation(
    Object inputBean,
    Class targetDomainObjectClass,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence,
    TransactionToken transactionToken){
        
        Object result=executeOperation(inputBean,targetDomainObjectClass,opConfig,sqlPersistence,transactionToken);
        
        if (result instanceof ArrayList){
            return (ArrayList)result;
        }
        else{
            ArrayList list=new ArrayList();
            if (result!=null) list.add(result);
            return list;
        }
        
    }
    
    /**
     * Loads and executes a SQL PreparedStatment and, if appropriate, extracts data from a ResultSet to return to caller, without
     * transactional support
     *
     * @param inputBean Contains input data the query will based on.
     * @param targetDomainObjectClass Used to generate newInstance()'s of a specific DomainObject to hold the results set data.
     * @param operationConfig Contains the data from the XML file that will be used to build a prepared statement.
     * @param sqlPersistence Contains the connection config info needed to make a database connection.
     * @param transactionToken Will be used to maintain an existing or generate a new transaction.
     * @return either a single DomainObject, a Collection of DomainObjects, or null.
     */
    public static Object executeOperation(
    Object inputBean,
    Class targetDomainObjectClass,
    RdbmsDAOOperationConfig opConfig,
    SqlPersistence sqlPersistence){
        
        return executeOperation(inputBean, targetDomainObjectClass, opConfig, sqlPersistence, null);
    }
    
    /**
     * Loads and executes a SQL PreparedStatment and, if appropriate, extracts data from a ResultSet to return to caller, with 
     * transactional support
     *
     * @param inputBean Contains input data the query will based on.
     * @param targetDomainObjectClass Used to generate newInstance()'s of a specific DomainObject to hold the results set data.
     * @param operationConfig Contains the data from the XML file that will be used to build a prepared statement.
     * @param sqlPersistence Contains the connection config info needed to make a database connection.
     * @param transactionToken Will be used to maintain an existing or generate a new transaction.
     * @return either a single DomainObject, an Collection of DomainObjects, or null.
     */
    public static Object executeOperation(
    Object inputBean,
    Class targetDomainObjectClass,
    RdbmsDAOOperationConfig operationConfig,
    SqlPersistence sqlPersistence,
    TransactionToken transactionToken){
        
        //1. get a connection based on the info in sqlPersistence and transactionToken
        //      1a. determine if this is part of an existing transaction
        //      1b. get the connection from the existing transaction instance or;
        //      1c. create a new, one-off transaction
        //2. create and load the prepared statement from the inputBean and operationConfig
        //3. execute the SQL statement
        //4. extract query results and build return object
        //5. returns either a DomainObject, an Collection of DomainObjects, or null
        
        //validate input data
        if (operationConfig==null) throw new IllegalArgumentException("RdbmsDAOOperationConfig specified was null!");
        
        String preparedStatementText=operationConfig.getPreparedStatement();
        if (preparedStatementText==null) {
            throw new IllegalArgumentException("Prepared Statement in RdbmsDAOOperationConfig was null!");
        }
        
        //get the connection - either an existing connection referenced by the
        //transaction token or a brand new connection
        Connection opConnection=createConnection(transactionToken,sqlPersistence);
        
        if (opConnection==null) {
            throw new PersistenceException("Unable to get Sql Connection - (opConnection==null).");
        }
        String opLabel = null;
        
        try{
            //build and load the PreparedStatement
            PreparedStatement opPreparedStatement = opConnection.prepareStatement(preparedStatementText);
            loadPreparedStatement(opPreparedStatement,inputBean,operationConfig);
            
            //are we expecting any query results?
            boolean noResultSet =
            operationConfig.getBeanQueryResultMappings()==null ||
            operationConfig.getBeanQueryResultMappings().isEmpty();
            
            //execute the prepared statement
            if (noResultSet){
                opLabel = "Update";
                opPreparedStatement.executeUpdate();
            }
            else{
                //execute the query
                opLabel = "Query";
                ResultSet opResultSet=opPreparedStatement.executeQuery();
                return createOperationResults(targetDomainObjectClass,opResultSet,operationConfig);
            }
        }
        catch (SQLException e){
            StringBuffer connectionInfo = new StringBuffer();
            connectionInfo.append("Hostname='");
            connectionInfo.append(sqlPersistence.getHostname());
            connectionInfo.append("' database='");
            connectionInfo.append(sqlPersistence.getDatabaseName());
            connectionInfo.append("' UserId='");
            connectionInfo.append(sqlPersistence.getUserId());
            connectionInfo.append("' Key='");
            connectionInfo.append(sqlPersistence.getKey());
            connectionInfo.append("' PersistenceTechnologyKey='");
            connectionInfo.append(sqlPersistence.getPersistenceTechnologyKey());
            connectionInfo.append("'");
            
            throw new PersistenceException("SQLException#" + e.getErrorCode() + 
                  " caught while executing operation:" + opLabel + " with \npreparedStatementText='" + 
                  preparedStatementText + "'  \nconnectionInfo=" + connectionInfo + 
                  "'\ncaught SQLException \ne.getClass()=[" + e.getClass() + "] e.getMessage()=[" 
                  + e.getMessage() + "]", e);
        }
        finally{
            try{
                if (transactionToken!=null){
                    //close the connection if we're not in a transaction
                    opConnection.close();
                }
            }
            catch (Throwable t){} //need to swallow this?
        }
        return null;
    }
    
    
    private static Connection createConnection(TransactionToken transactionToken, SqlPersistence sqlPersistence){
        
        Connection opConnection=null;
        boolean inTransaction = false;
        
        //prepare the Connection
        if (transactionToken!=null){
            //prepare Transactional Connection
            inTransaction=true;
            TransactionInstance transactionInstance= TransactionCoordinator.getTransactionInstance(transactionToken);
            
            //are we starting a new transaction?
            //i.e. is this the first operation executed in a transaction?
            if (transactionInstance==null){
                opConnection=sqlPersistence.getConnection();
                RdbmsTransactionInstance rdbmsTransactionInstance=new RdbmsTransactionInstance(opConnection);
                TransactionCoordinator.startTransaction(transactionToken,rdbmsTransactionInstance);
            }
            //we're already in an existing transaction... get the existing connection
            else{
                if (RdbmsTransactionInstance.class.isAssignableFrom(transactionInstance.getClass())){
                    opConnection=((RdbmsTransactionInstance)transactionInstance).getConnection();
                    if (opConnection==null) {
                        throw new PersistenceException("Existing transaction detected," + 
                                           " but retrieved a null Connection! (expected an active connection)");
                    }
                }
                else{
                    throw new PersistenceException("Incompatible transaction instance retrieved." +  
                                                   "Expected RdbmsTransactionInstance; instead got " + 
                                                    transactionInstance.getClass().getName());
                }
            }
        }
        else{
            //prepare non-Transactional Connection
            opConnection=sqlPersistence.getConnection();
            inTransaction=false;
        }
        
        return opConnection;
        
    }
    
    private static Object createOperationResults(
    Class targetDomainObjectClass,
    ResultSet opResultSet,
    RdbmsDAOOperationConfig operationConfig) throws SQLException{
        
        //extract data from the ResultSet
        //create a new domain object instance to return the query results in, 
        //or an Collection of results as appropriate
        ArrayList resultList=new ArrayList();
        DomainObject domainObject=null;
        
        //move ResultSet data into an Collection of DomainObjects
        while(opResultSet.next()){
            try{
                // if we've been given an interface use the factory to get an instance
                // otherwise just get an instance from the class object
                if (true == targetDomainObjectClass.isInterface()) {
                	domainObject=(DomainObject)DomainObjectFactory.createDOInstance(targetDomainObjectClass);
		} else {
			domainObject=(DomainObject)targetDomainObjectClass.newInstance();
		}
            } catch (InstantiationException e) {
                throw new PersistenceException("Exception caught while creating the instance of the domain object '" 
                                          + targetDomainObjectClass.getName()+"'.\ne.getClass()=[" + e.getClass() 
                                          + "] e.getMessage()=[" + e.getMessage() + "]", e);
            } catch (IllegalAccessException e) {
                throw new PersistenceException("Exception caught while creating the instance of the domain object '" 
                                          + targetDomainObjectClass.getName()+"'.\ne.getClass()=[" + e.getClass() 
                                          + "] e.getMessage()=[" + e.getMessage() + "]", e);
            } catch (RuntimeException e) {
                throw new PersistenceException("Exception caught while creating the instance of the domain object '" 
                                          + targetDomainObjectClass.getName()+"'.\ne.getClass()=[" + e.getClass() 
                                          + "] e.getMessage()=[" + e.getMessage() + "]", e);

            } 
            updateBeanFromResultSet(domainObject,opResultSet,operationConfig);
            resultList.add(domainObject);
        }
        //return the results as an ArrayList, DomainObject, or null
        //return the list if there are more than 1 records
        if (resultList.size()>1)
            return resultList;
        //return just the domain object if only one record was retrieved
        else if (resultList.size()==1)
            return domainObject;
        else
            //nothing was retrieved... return null
            return null;
        
    }
    
    
    private static void updateBeanFromResultSet(
    Object targetBean,
    ResultSet resultSet,
    RdbmsDAOOperationConfig opConfig){
        
        ArrayList beanQueryResultMappings = opConfig.getBeanQueryResultMappings();
        
        ResultSetToBeanMapper mapper = 
            (ResultSetToBeanMapper)ObjectUtil.createObjectFromConfig(PersistenceConstants.PERSISTENCE_OPTIONS_KEY,
                                                                     PersistenceConstants.RESULTSET_TO_BEAN_MAPPER_KEY);
        mapper.setTargetBean(targetBean);
        mapper.setSourceResultSet(resultSet);
        
        
        try{
            ResultSetMetaData rsMeta = resultSet.getMetaData();
            
            for (Iterator i=beanQueryResultMappings.iterator();i.hasNext();){
                QueryResultToBeanMapItem queryResultToBeanConfig=(QueryResultToBeanMapItem)i.next();

                mapper.setBeanProperty(queryResultToBeanConfig);
            }
        }
        catch (SQLException e){
            throw new PersistenceException("SQLException caught while updating domain object.\ne.getClass()=[" + 
                                            e.getClass() + "] e.getMessage()=[" + e.getMessage() + "]", e);
        }
    }

    private static void loadPreparedStatement(
    PreparedStatement preparedStatement,
    Object sourceBean,
    RdbmsDAOOperationConfig opConfig){
        
        //get the mapper object
        BeanToSqlParameterMapper mapper = 
           (BeanToSqlParameterMapper)ObjectUtil.createObjectFromConfig(PersistenceConstants.PERSISTENCE_OPTIONS_KEY,
                                                                       PersistenceConstants.BEAN_TO_SQL_PARAMETER_MAPPER_KEY);
       
        mapper.setTargetPreparedStatement(preparedStatement);
        
        // not every query needs a parameter 
        if (null != opConfig.getBeanParameterMappings() ) { 
	    mapper.setSourceBean(sourceBean);
           for (Iterator i=opConfig.getBeanParameterMappings().iterator();i.hasNext();){
            
               OrdinalParameterToBeanMapItem paramToBeanConfig = (OrdinalParameterToBeanMapItem)i.next();
               mapper.writeParameterValue(paramToBeanConfig);
            }
        }
    }
}
