// (c) Copyright 2001-2002 iwombat.com all rights reserved

package com.iwombat.util;



import java.lang.reflect.*;

import java.util.*;

/**
 * Class PrototypeUtil
 * Description: Utility class to support the GOF prototype pattern through bean reflection
 *
 * @author  $Author: brentp $
 * @version $Revision: 1.3 $
 * 
 * <br>
 * <dl><dt><b>Date:</b><dd>$Date: 2002/11/18 16:04:43 $</dd></dl>
 */

public class BeanUtil {
    static final String getPrefix = "get";
    static final String setPrefix = "set";

    
    protected BeanUtil() {}
    /**
     *
     * Method synchronizeData
     * Description: This method reflects on the sharedSuper class and gathers all
     * public accessors (getters). It then uses the accessors on the source to provide parameters
     * to the mutators (setters) on the target class.
     *
     * @param target object
     * @param source object
     * @param shared superclass or interface
     * @return the target object
     * @throws IllegalArgumentExecption if one of the follwing conditions is true:
     * <ul><li>target or source is not assignable as the sharedSuper. 
     *     <li>the super class has no public accessors
     *     <li>a security violation occurs while performing reflection on any class
     * </ul>
     * @throws RuntimeException if any method.invoke() has issues.
     */
    
    public static Object synchronizeData(Object target, Object source, Class sharedSuper) {
        
        String  name           = null;
        Object  value          = null;
        
        Class   sourceClass    = source.getClass();
        Class   targetClass    = target.getClass();
        
        // make sure both classes implement the interface
        if ( !sharedSuper.isAssignableFrom(targetClass) ) {  
            throw new IllegalArgumentException("target:" + targetClass.getName() + 
                                                "does not implement specified interface:" + sharedSuper.getName() );
        }       
        if ( !sharedSuper.isAssignableFrom(sourceClass) ) {
            throw new IllegalArgumentException("source:" + sourceClass.getName() + 
                                                "does not implement specified interface:" + sharedSuper.getName() );
        }
        
        try {
            ArrayList accessors = getAccessors(sharedSuper);
            if (accessors.isEmpty()) {
                throw new IllegalArgumentException("interface has no public member methods");
            }
            
            //invoke methods
            Iterator iter = accessors.iterator();
            while (iter.hasNext()) {
                 setValue(target, source, (Method)iter.next());
            }
            
        } catch (SecurityException e) {
            throw new IllegalArgumentException("Inadequate permissions on target or source" + e.getMessage());           
        }
        
        return target;
        
    }

    /**
     * This method gets all the public accessor methods (getters) and builds an
     * array list.
     *
     */
    private static ArrayList getAccessors(Class theClass) {
        ArrayList accessors = new ArrayList();
        Method[] publicMethods = theClass.getMethods();
        
        for (int i=0; i<publicMethods.length; i++) {
            // if we find the getPrefix at position 0 this must be a getter
            if (0 == publicMethods[i].getName().indexOf(getPrefix)) {
                accessors.add(publicMethods[i]);
            }
        }
        return accessors;
    }

    
    /**
     * Given an accessor method, construct the mutator and use one to provide 
     * parameters to invoke the other.
     *
     */
    private static void setValue(Object target, Object source, Method accessor) {
        String mutatorName = setPrefix + accessor.getName().substring(setPrefix.length());
        
        Class   sourceClass    = source.getClass();
        Class   targetClass    = target.getClass();
        
        Method sourceMethod = accessor;
        Method targetMethod = null;
        Class[] resultTypeArray = new Class[1];
        resultTypeArray[0] = accessor.getReturnType();
        try {
            targetMethod = targetClass.getMethod(mutatorName, resultTypeArray);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException("No method in target object named:" + accessor.getName() +
            " with parameter:" + resultTypeArray[0].getClass().getName() );
        }
        
        
        Object resultArray[] = new Object[1];
        
        // perform the reflextion operations, trap all the bloody exceptions possible.
        try {
            try {
                resultArray[0] = sourceMethod.invoke(source, null);
            } catch (IllegalAccessException e) {
                throw new IllegalArgumentException("No access to method in source object named:"
                + sourceMethod.getName() + "\n" + e.getMessage());
            }
            
            try {
                targetMethod.invoke(target, resultArray);
            } catch (IllegalAccessException e) {
                throw new IllegalArgumentException("No access to method in target object named:"
                + targetMethod.getName() + "\n" + e.getMessage());
            }
        } catch (InvocationTargetException e) {
            throw new RuntimeException("Exception occured in reflexion operations:" + e.getMessage());           
        }
    }

   public static String createMutatorMethodName(String beanPropertyName){
        
        StringBuffer setMethod=new StringBuffer();
        setMethod.append(setPrefix);
        setMethod.append(beanPropertyName.substring(0,1).toUpperCase());
        setMethod.append(beanPropertyName.substring(1));
        
        return setMethod.toString();
    }
    
    public static String createAccessorMethodName(String beanPropertyName){
        
        StringBuffer getMethod=new StringBuffer();
        getMethod.append(getPrefix);
        getMethod.append(beanPropertyName.substring(0,1).toUpperCase());
        getMethod.append(beanPropertyName.substring(1));
        
        return getMethod.toString();
    }    
}
