// (c) Copyright 2001-2002 iwombat.com all rights reserved

package com.iwombat.util;

/**
 * An object wrapper for byte arrays, in the style of first class object representations of primitives such as Long or Float 
 *
 * @author  $Author: bob $
 * @version $Revision: 1.4 $
 * 
 * <br>
 * <dt><b>Date:</b><dd>$Date: 2003/09/04 18:37:04 $</dd>
 */
public class ByteArray {
    
    private byte[] byteArrayValue;
    
    /**
     * Creates a new instance of ByteArray using the specified byte array primitive
     */
    public ByteArray(byte[] byteArrayValue) {
        this.setByteArrayValue(byteArrayValue);
    }
    
    public ByteArray(byte singleByte) {
        byte[] onsie = new byte[1];
        onsie[0] = singleByte;
        this.setByteArrayValue(onsie);
    }

    /**
     * Creates a new instance of ByteArray using the specified hex string
     */
    public ByteArray(String hexString){
     
        this.setByteArrayValue(HexStringUtil.bytesFromHexString(hexString));
    }
    
    public ByteArray(float floatValue) {
        byteArrayValue = toByteArray(floatValue);
 
    }
    
    /**
     * returns the byteArrayValue
     *
     * @return the byte array value
     */
    public byte[] byteArrayValue(){

        return this.getByteArrayValue();
    }
    
    /**
     * Compare two ByteArray objects.
     *
     * @return true if the ByteArray objects are the same, false otherwise. 
     */
    public boolean equals(Object obj){   
        if (!(obj instanceof ByteArray)) return false;       
        return ByteArrayUtil.sameValue(this.byteArrayValue(),((ByteArray)obj).byteArrayValue());        
    }
   
    /**
     * Get the internal byte array
     *
     * @return the internal byte array. 
     */
    protected byte[] getByteArrayValue(){

        return this.byteArrayValue;
    }
    
    /**
     * Set the internal byte array
     *
     * @param byteArrayValue - set the internal byte array. 
     */
    protected void setByteArrayValue(byte[] byteArrayValue){
        
        this.byteArrayValue=byteArrayValue;
    }
    
    /**
     * Returns the number of bytes in the internal byte[]. 
     *
     * @return the number of bytes in the byte array. 
     */
    public int getLength(){
      int n = 0;
      if (null != byteArrayValue)
        n = byteArrayValue.length;
      return n;
    }
    
    /**
     * Converts a float to a byte array
     *
     * @param floatValue - the value of the float
     * @returns byte array for float in MSB order 
     */
    private static final short BYTE_MASK = 0xff;
    private static final short BYTE_LEN = 8; // 8-bits in a byte
    private static final short FLOAT_LEN = 4;
    private static final short DOUBLE_LEN = 8;
    private static final short SHORT_LEN = 2;
    private static final short INT_LEN = 4;
    private static final short LONG_LEN = 8;
    
    public static byte[] toByteArray(float floatValue) {
        byte[] tempArray = new byte[FLOAT_LEN];
        int i = Float.floatToIntBits(floatValue);
        for (int pos=FLOAT_LEN-1; pos>=0; pos--) {
            tempArray[pos] = (byte)(i & BYTE_MASK);
            i >>>= BYTE_LEN;
        }
        return tempArray;
    }
    
    public static byte[] toByteArray(double doubleValue) {
        byte[] tempArray = new byte[DOUBLE_LEN];
        long i = Double.doubleToLongBits(doubleValue);
        for (int pos=DOUBLE_LEN-1; pos>=0; pos--) {
            tempArray[pos] = (byte)(i & BYTE_MASK);
            i >>>= BYTE_LEN;
        }
        return tempArray;
    }
    public static byte[] toByteArray(short shortValue) {
        byte[] tempArray = new byte[SHORT_LEN];
        for (int pos=SHORT_LEN-1; pos>=0; pos--) {
            tempArray[pos] = (byte)(shortValue & BYTE_MASK);
            shortValue >>>= BYTE_LEN;
        }
        return tempArray;
    }
    public static byte[] toByteArray(int intValue) {
        byte[] tempArray = new byte[INT_LEN];
        for (int pos=INT_LEN-1; pos>=0; pos--) {
            tempArray[pos] = (byte)(intValue & BYTE_MASK);
            intValue >>>= BYTE_LEN;
        }
        return tempArray;
    }
    
    public static byte[] toByteArray(long longValue) {
        byte[] tempArray = new byte[LONG_LEN];
        for (int pos=LONG_LEN-1; pos>=0; pos--) {
            tempArray[pos] = (byte)(longValue & BYTE_MASK);
            longValue >>>= BYTE_LEN;
        }
        return tempArray;
    }
    
    /**
     * Returns a byte array of appropriate length. This static method clips from
     * lest significant byte (end of array) to most significant byte (beginning).
     * This may result in loss of precision.
     *
     * @param byte array
     * @param length of desired clipped byte array
     * @return new cliped byte array 
     */
    public static byte[] clipMSBArray(byte[] array, int length) {
        byte[] newArray = new byte[length];
        int offset = array.length - length;
        if (offset < 0) {
            throw new IllegalArgumentException("invalid length specified");
        }
        
        for (int i=length-1; i>=0; i--) {
            newArray[i] = array[i+offset];
        }
        return newArray;
    }
    
    public static byte[] chompArray(byte[] array, int length) {
        byte[] newArray = new byte[length];
        
        for (int i=0; i<length; i++) {
            newArray[i] = array[i];
        }
        return newArray;
    }
    
    public static long bytesToLong(byte[] rawLong) {
        long value = 0;
        for (int i=0; i < rawLong.length; i++) {
            value <<= BYTE_LEN;
            value |= ( rawLong[i] & BYTE_MASK );
        }
        
        return value;
    }

    public static boolean sameValue(byte[] baFirst, byte[] baSecond) {

        int baFirstLength  = baFirst.length;
        int baSecondLength = baSecond.length;

        if (baFirstLength != baSecondLength) {
            return false;
        }

        for (int i = 0; i < baFirstLength; i++) {
            if (baFirst[i] != baSecond[i]) {
                return false;
            }
        }

        return true;
    }
   
 }

