// (c) Copyright 2001-2002 iwombat.com all rights reserved

package com.iwombat.util;


import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;

import java.io.Serializable;

/**
 * The SmartDetails class is a helper class to maintain "details" records for
 * a database table. The name/value pairs of a regular HashMap are retained with
 * an additional flag for what's been changed. This is to reduce the number of
 * times a database record is actually added,changed, or deleted.
 * <P>
 * <B>Note:</B> Like other default members of java.util.Collections, this
 * object is <B>note</B> thread-safe.
 * <P>
 * 
 * @author  $Author: bob $
 * @version $Revision: 1.3 $
 * 
 * <br>
 * <dl><dt><b>Date:</b><dd>$Date: 2002/10/01 17:54:16 $</dd></dl>
* @see <{Map}>
*/
public class SmartHash implements Map,Serializable {
    /**
    * Flag to indicate a new element since initial load or cleanup.
    */
    private static final int ADDED = 1;
    /**
    * Flag to indicate a deleted element since initial load or cleanup.
    */
    private static final int REMOVED = 2;
    /**
    * Flag to indicate a changed element since initial load or cleanup.
    */
    private static final int CHANGED = 3;
    /**
    * Flag to indicate an unchanged element since initial load or cleanup.
    */
    private static final int ORIGINAL = 0;
    /**
    * Actual data content.
    */
    private HashMap data;

    /**
    * Default constructor. Creates initial data store.
    */
    public SmartHash() {
        data = new HashMap();
    }

    /**
    * Constructor that takes another SmartHash object and "clones" the data.
    * This includes the "state" of the data within the hash.
    * @param hash the object to create a duplicate of
    */
    public SmartHash(SmartHash hash) {
        data = new HashMap(hash.data);
    }

    //-------------------------------------------------------------------------
    // Methods required by the Map interface.
    //-------------------------------------------------------------------------

    /**
    * Remove all name/value pairs. <B>Note:</B> this does not acutally remove
    * the pairs from the hash: it only marks them as "deleted."
    */
    public void clear() {
        Iterator it = data.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            remove(key);
        }
    }

    /**
    * Determine if the specified name is a key in the map.
    * @param name the name of the key to test
    * @return true if the key exists in the map, false otherwise
    */
    public boolean containsKey(Object name) {
        return data.containsKey(name);
    }

    /**
    * Determine if the specified value is contained within the map.
    * @param value the value to test for
    * @return true if the value exists in the map, false otherwise; returns
    * false for nulls
    */
    public boolean containsValue(Object value) {
        if (value == null)
            return false;
        Iterator it = data.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            Detail row = (Detail)data.get(key);
            if (row.value == null)
                continue;
            if (row.value.equals(value))
                return true;
        }
        return false;
    }

    /**
    * Returns a set view of the mappings contained in this map. The Map.Entry
    * objects will actually represent a HashMap of the name/value pairs and not
    * the flags.
    * @return a set of Map.Entry objects
    */
    public Set entrySet() {
        HashMap hash = new HashMap();
        Iterator it = data.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            Detail row = (Detail)data.get(key);
            hash.put(key,row.value);
        }
        return hash.entrySet();
    }

    /**
    * Compares the specified object with this map for equality.
    * @param o object to be compared for equality with this map
    * @return true if the specified object is equal to this map
    */
    public boolean equals(Object o) {
        if (o instanceof SmartHash)
            return data.equals(((SmartHash)o).data);
        return false;
    }

    /**
    * Method for accessing values.
    * <P>
    * <B>Note:</B> "removed" values are not returned
    * @param name the name or key of the item
    * @return the value associated with the name
    */
    public Object get(Object name) {
        Detail row = (Detail)data.get(name);
        if (row == null)
            return null;
        if (row.flag == REMOVED)
            return null;
        return row.value;
    }

    /**
    * Get the hash for this map.
    * @return the hash code for this map
    */
    public int hashCode() {
        return data.hashCode();
    }

    /**
    * Determine if there's any data in the map.
    * @return true if the map is empty, false if there's data in it
    */
    public boolean isEmpty() {
        return data.isEmpty();
    }

    /**
    * Returns a set view of the keys contained in this map.
    * @return a set view of the keys contained in this map
    */
    public Set keySet() {
        return data.keySet();
    }

    /**
    * Method for adding/changing values. Compares the current value with the
    * new one to determine whether it has actually changed. New names (keys)
    * are simply added. Null keys are ignored. Null values are stored as "place
    * holders."
    * @param name the name or key of the item
    * @param value the value to be associated with this name/key
    * @return previous value associated with specified key, or null if there
    * was no mapping for key
    */
    public Object put(Object name,Object value) {
        if (name == null)
            return null;

        Detail row = (Detail)data.get(name);
        Object rtn = null;

        // no previous entry for this key - make a new entry
        if (row == null) {
            row = new Detail();
            row.flag = ADDED;
        }

        // previous entry is empty
        else if (row.value == null) {
            if (value != null)
                row.flag = CHANGED;
        }

        // values not equivalent: here we're trusting to the object type of the
        // value to provide the equals method
        else if (! row.value.equals(value)) {
            row.flag = CHANGED;
            rtn = row.value;
        }

        // values are equivalent, but previously removed: reset to non-changed
        // state
        else if (row.flag == REMOVED) {
            row.flag = ORIGINAL;
            rtn = row.value;
        }

        else
            rtn = row.value;

        row.value = value;
        data.put(name,row);
        return rtn;
    }

    /**
    * Copies all of the mappings from the specified map to this map. Data
    * states are created on the fly as above. Null keys are ignored.
    * @param map mappings to be stored in this map
    */
    public void putAll(Map map) {
        Iterator it = map.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            put(key,map.get(key));
        }
    }

    /**
    * Marks a name/key for removal.
    * @param name the name or key of the item to "remove"
    * @return the value of the removed mapping
    */
    public Object remove(Object name) {
        Detail row = (Detail)data.get(name);
        if (row == null)
            return null;
        row.flag = REMOVED;
        return row.value;
    }

    /**
    * Get the size of the hash.
    * @return the number of entires in the map
    */
    public int size() {
        return data.size();
    }

    /**
    * Retrieve a list of the values contained within the map.
    * @return the list of values that have not been removed
    */
    public Collection values() {
        ArrayList v = new ArrayList();
        Iterator it = data.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            Detail row = (Detail)data.get(key);
            if (row.value != null && row.flag != REMOVED)
                v.add(row.value);
        }
        return v;
    }

    //-------------------------------------------------------------------------
    // New methods for this class
    //-------------------------------------------------------------------------

    /**
    * Retrieve a value, regardless of it's state.
    * @param name the name or key of the item
    * @return the value associated with the name
    */
    public Object getAnyway(Object name) {
        Detail row = (Detail)data.get(name);
        if (row == null)
            return null;
        return row.value;
    }

    /**
    * Initial load of a name/value pair. This is useful when
    * initializing the data store.
    * @param name the name or key of the item
    * @param value the value to be associated with this name/key
    */
    public void load(Object name,Object value) {
        Detail row = new Detail();
        row.value = value;
        row.flag = ORIGINAL;
        data.put(name,row);
    }

    /**
    * Initial load of data. This method assumes that the data is "fresh" and
    * unchanged and will reflect that in the data "state."
    * @param map the data to initialize
    */
    public void loadAll(Map map) {
        Iterator it = map.keySet().iterator();
        while (it.hasNext()) {
            Object key = it.next();
            load(key,map.get(key));
        }
    }

    /**
    * Clears all CHANGED/ADDED tags and removes all elements marked REMOVED.
    * This method should be used after the data has been moved to persistent
    * storage.
    */
    public void cleanUp() {
        Iterator it = data.keySet().iterator();
        while (it.hasNext()) {
            String key = it.next().toString();
            Detail row = (Detail)data.get(key);
            if (row.flag != REMOVED)
                row.flag = ORIGINAL;
            else
                it.remove();
        }
    }

    /**
    * Determine if any part of the names/values have changed.
    */
    public boolean isChanged() {
        boolean changed = false;
        Iterator it = data.keySet().iterator();
        while (it.hasNext() && !changed) {
            String key = it.next().toString();
            Detail row = (Detail)data.get(key);
            changed = (row.flag != ORIGINAL);
        }
        return changed;
    }

    /**
    * Determine if a value has been changed.
    * @param name the name or key of the item
    */
    public boolean isChanged(Object name) {
        return checkFlag(name,CHANGED);
    }

    /**
    * Determine if a value has been added.
    * @param name the name or key of the item
    */
    public boolean isNew(Object name) {
        return checkFlag(name,ADDED);
    }

    /**
    * Determine if a value has been deleted.
    * @param name the name or key of the item
    */
    public boolean isRemoved(Object name) {
        return checkFlag(name,REMOVED);
    }

    /**
    * Method to check vs. a specific flag.
    * @param name the name or key of the item
    * @param flag the flag to check a value against
    */
    private boolean checkFlag(Object name,int flag) {
        Detail row = (Detail)data.get(name);
        return row.flag == flag;
    }

    /**
    * Helper class: represents the value and the state of the data.
    */
    class Detail implements Serializable {
        Object value;
        int flag;
    }
}

